/*

Copyright (c) 2010, Embedded Adventures, www.embeddedadventures.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of Embedded Adventures nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
THE POSSIBILITY OF SUCH DAMAGE.

Contact us at admin@embeddedadventures.com

*/


/**
	\file m41t81s.h
	\brief Routines for communicating with the m41t81s real time clock
*/


#ifndef __m41t81s_h
#define __m41t81s_h defined

#include "i2c.h"
#include "pic_utils.h"

// frequency of square wave output

#define rtc_sqw_freq_32768Hz	0b00000001
// 16384Hz not available
#define rtc_sqw_freq_8192Hz		0b00000010
#define rtc_sqw_freq_4096Hz		0b00000011
#define rtc_sqw_freq_2048Hz		0b00000100
#define rtc_sqw_freq_1024Hz		0b00000101
#define rtc_sqw_freq_512Hz		0b00000110
#define rtc_sqw_freq_256Hz		0b00000111
#define rtc_sqw_freq_128Hz		0b00001000
#define rtc_sqw_freq_64Hz		0b00001001
#define rtc_sqw_freq_32Hz		0b00001010
#define rtc_sqw_freq_16Hz		0b00001011
#define rtc_sqw_freq_8Hz		0b00001100
#define rtc_sqw_freq_4Hz		0b00001101
#define rtc_sqw_freq_2Hz		0b00001110
#define rtc_sqw_freq_1Hz		0b00001111


/** The m41t81s device address */
#define m41t81s_device_addr  0xD0

/** m41t81s tenths and hundreths of seconds register */
#define m41t81s_part_seconds_reg 0x00
/** m41t81s seconds register (D7=ST)*/
#define m41t81s_seconds_reg 0x01
/** m41t81s minutes register */
#define m41t81s_minutes_reg 0x02
/** m41t81s hours register (D7=CEB, D6=CB) */
#define m41t81s_hours_reg   0x03
/** m41t81s day of week register */
#define m41t81s_dow_reg     0x04
/** m41t81s date in month register */
#define m41t81s_date_reg    0x05
/** m41t81s month register */
#define m41t81s_month_reg   0x06
/** m41t81s year register */
#define m41t81s_year_reg  0x07
/** m41t81s calibration register  (D7=OUT, D6=FT, D5=S D4-D0=Calibration) */
#define m41t81s_calibration_reg 0x08
/** m41t81s watchdog register  (D7=OFIE, D6-D2=BMB, D1-D0=RB */
#define m41t81s_watchdog_reg 0x09
/** m41t81s alarm month register  (D7=AFE, D6=SQWE, D5=ABE, D4=AL 10M, D3-D0=Alarm month)*/
#define m41t81s_alarm_month_reg 0x0A
/** m41t81s alarm date register  (D7=RPT4, D6=RPT5, D5-D4=ABE, D4=AL 10M, D3-D0=Alarm month)*/
#define m41t81s_alarm_date_reg 0x0B
/** m41t81s alarm hour register  (D7=RPT3, D6=HT, D5-D4=Alarm 10 Hour, D3-D0=Alarm Hour)*/
#define m41t81s_alarm_hour_reg 0x0C
/** m41t81s alarm min register  (D7=RPT2, D6-D4=Alarm 10 Minutes, D3-D0=Alarm Minutes)*/
#define m41t81s_alarm_min_reg 0x0D
/** m41t81s alarm seconds register  (D7=RPT1, D6-D4=Alarm 10 Seconds, D3-D0=Alarm Seconds)*/
#define m41t81s_alarm_seconds_reg 0x0E
/** m41t81s flags register  (D7=WDF, D6=AF, D5=0, D4=BL, D3=0, D2=OF, D1=0, D0=0 -D4=Alarm 10 Seconds, D3-D0=Alarm Seconds)*/
#define m41t81s_flags_reg 0x0F
/** m41t81s reserved register */
#define m41t81s_reserved1_reg 0x10
/** m41t81s reserved register */
#define m41t81s_reserved2_reg 0x11
/** m41t81s reserved register */
#define m41t81s_reserved3_reg 0x12
/** m41t81s SQW register (D7-D4=RS3-RS0)*/
#define m41t81s_sqw_reg 0x13

/*
0 = Must be set to '0'
ABE = Alarm in Battery Back-up Mode Enable Bit
AF = Alarm Flag (Read only)
AFE = Alarm Flag Enable Flag
BL = Battery Low Bit
BMB0-BMB4 = Watchdog Multiplier Bits
CB = Century Bit
CEB = Century Enable Bit
FT = Frequency Test Bit
HT = Halt Update Bit
OF = Oscillator Fail Flag
OFIE = Oscillator Fail Interrupt Enable
OUT = Output level
RB0-RB1 = Watchdog Resolution Bits
RPT1-RPT5 = Alarm Repeat Mode Bits
RS0-RS3 = SQW Frequency
S = Sign Bit
SQWE = Square Wave Enable
ST = Stop Bit
WDF = Watchdog Flag (Read only)
*/









/**
	\brief Get the decoded minutes register from the m41t81s
	
	Returns the number of minutes past the hour from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range - 0 through 59
*/	
uns8 rtc_get_minutes();


/**
	\brief Get the decoded hours register from the m41t81s
	
	Returns hour from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.  These routines
	assume the m41t81s is running in 24 hour mode.
	Range - 0 through 23


*/
uns8 rtc_get_hours();


/**
	\brief Get the decoded seconds register from the m41t81s
	
	Returns seconds from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range - 0 through 59
	
*/	
uns8 rtc_get_seconds();

/**
	\brief Get the day register from the m41t81s
	
	Returns the day of the week from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range - 1 through 7
*/
uns8 rtc_get_dow();


/**
	\brief Get the date register from the m41t81s
	
	Returns the date in month from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range 1 through 28/29/30/31 depending on month
*/
uns8 rtc_get_date();


/**
	\brief Get the month register from the m41t81s
	
	Returns the month of the year from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range 1 through 12
*/	
uns8 rtc_get_month();


/**
	\brief Get the year register from the m41t81s
	
	Returns the year from the m41t81s.  The result is
	coverted to decimal from BCD and is ready to use.
	Range 0 through 99
*/	
uns8 rtc_get_year();


/**
	\brief Set the year register from the m41t81s
	
	Changes the year in the m41t81s.
*/	
void rtc_set_year(uns8 year);

/**
	\brief Set the minutes register from the m41t81s
	
	Changes the minutes in the m41t81s.
	
	\param seconds Value to set minutes to

*/	
void rtc_set_minutes(uns8 minutes);


/**
	\brief Set the day of the week register from the m41t81s
	
	Changes the day of the week in the m41t81s.
	
	\param seconds Value to set day to

*/	
void rtc_set_day(uns8 day);
/**
	\brief Set the date register from the m41t81s
	
	Changes the date in the m41t81s.

	\param seconds Value to set date to

*/	
void rtc_set_date(uns8 date);

/**
	\brief Set the seconds register in the m41t81s
	
	Changes the seconds in the m41t81s.
	
	\param seconds Value to set seconds to
*/	
void rtc_set_seconds(uns8 seconds);

/**
	\brief Set the hours register in the m41t81s
	
	Changes the hours in the m41t81s.  Forces the m41t81s into
	24 hour mode.
*/	
void rtc_set_hours(uns8 hours);

/**
	\brief Set the month register in the m41t81s
	
	Changes the month in the m41t81s.
*/	
void rtc_set_month(uns8 month);

/**
	\brief Set the config register in the m41t81s
	
	
	\param config Value to set the config register to
*/	
uns8 rtc_set_config(uns8 config);

/**
	\brief Stop the clock in the m41t81s
	
	Pauses time in the m41t81s
*/	
void rtc_stop_clock();


/**
	\brief Starts the clock in the m41t81s
	
	Resume time in the m41t81s. Also resumes the paused time
	that happens upon non-battery backup start up (which allows you
	to read the time before "resuming" so you know how long the clock
	has been running on battery back up).
	
	If you want to do this, read the time etc before calling rtc_start_clock();
*/	
void rtc_start_clock();

/**
	\brief Set the frequence of the square wave output pin
	
	Use one of the following self explanitory defines:
	
	rtc_sqw_freq_32768Hz
	rtc_sqw_freq_8192Hz
	rtc_sqw_freq_4096Hz
	rtc_sqw_freq_2048Hz
	rtc_sqw_freq_1024Hz
	rtc_sqw_freq_512Hz
	rtc_sqw_freq_256Hz
	rtc_sqw_freq_128Hz
	rtc_sqw_freq_64Hz
	rtc_sqw_freq_32Hz
	rtc_sqw_freq_16Hz
	rtc_sqw_freq_8Hz
	rtc_sqw_freq_4Hz
	rtc_sqw_freq_2Hz
	rtc_sqw_freq_1Hz

	Note that on the m41t81s 18384Hz is not available.
	
*/	
void rtc_set_sqw_freq(uns8 freq);

/**
	\brief Start pulsing on square wave output pin
	
	Outputs desired frequency on the SQW output pin. To set the frequency,
	see rtc_set_sqw_freq(uns8 freq);
	
*/	
void rtc_start_sqw_output();

/**
	\brief Stop pulsing on square wave output pin
	
	Stops square wave output
	
*/	
void rtc_stop_sqw_output();


/**
	\brief Setup ports and pins for use in the m41t81s
	
	Calls i2c_setup() to configure ports and pins ready for use 
*/	
void rtc_setup_io();
#define rtc_setup() rtc_setup_io()
uns8 rtc_get_register(uns8 reg);
void rtc_set_register(uns8 reg, uns8 data);

#endif
