/*

Copyright (c) 2010, Embedded Adventures, www.embeddedadventures.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of Embedded Adventures nor the names of its contributors
  may be used to endorse or promote products derived from this software
  without specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
THE POSSIBILITY OF SUCH DAMAGE.

Contact us at admin@embeddedadventures.com

*/
/** \file wpan.c	
    \brief Wireless Personal Area Network routines
*/

// Define this for debug output
#define debug_on


#include "wpan.h"
#include "pic_utils.h"
#include "pic_serial.h"
#include "debug.h"

// Not the neatest way to this but will do for now
uns8 wpan_rx_buffer[50];
uns8 wpan_rx_count;
uns8 receive_lost = 0;

uns8 pkt_received = 0;


void mrf24j40_receive_callback() {
	serial_putc('r');
	if (pkt_received) {
		receive_lost++;
		debug_str("<Lost receive>");
		mrf24j40_flush_receive_buffer();
	} else {
		pkt_received++;
		wpan_rx_count = mrf24j40_receive(&wpan_rx_buffer, sizeof(wpan_rx_buffer));
		
		debug_str("Rx: ");
		debug_int(wpan_rx_count);
		debug_str(" bytes ");
	}
	
	// this will get picked up in wpan_process	

}

void mrf24j40_transmit_callback(uns8 status,		// 0 = success
									   uns8 retries,	// retries to transmit
									   uns8 channel_busy) {	// 1 if fail due to channel busy
	debug_str(" <Tx: ");
	if (!status) {
		debug_str(" Good");
	} else {
		debug_str(" Fail");
		if (channel_busy) {
			debug_str(" (channel busy)");
		}	
	}
	debug_str(" Retries: ");
	debug_int(retries);
	debug_str(">\n");

	wpan_data_transmitted_callback(status, retries, channel_busy);

}


void wpan_print_address(wpan_address *addr) {
	
uns8 count;

	debug_str("\nPKT:\nSrc ");
	if (addr->source_address_type) { // ie, not 0
		serial_print_str(" PAN: ");
		serial_print_int_hex_16bit(addr->source_pan_id);
		
		switch (addr->source_address_type) {
			case WPAN_ADDR_TYPE_EXTENDED:
				serial_print_str(" EA: ");
				for (count = 7; count != 255; count--) {
					serial_print_int_hex(addr->source_ea[count]);
					serial_putc(' ');
				}
				break;
			case WPAN_ADDR_TYPE_SHORT:
				serial_print_str(" SA: ");
				serial_print_int_hex_16bit(addr->source_sa);
				serial_putc(' ');
				break;
		}			
	} else {
		serial_print_str(" (blank) ");
	}
	
	serial_print_str(" Dest ");
	
	if (addr->dest_address_type) { // ie, not 0
		serial_print_str(" PAN: ");
		serial_print_int_hex_16bit(addr->dest_pan_id);
		
		switch (addr->dest_address_type) {
			case WPAN_ADDR_TYPE_EXTENDED:
				serial_print_str(" EA: ");
				for (count = 7; count != 255; count--) {
					serial_print_int_hex(addr->dest_ea[count]);
					serial_putc(' ');
				}
				break;
			case WPAN_ADDR_TYPE_SHORT:
				serial_print_str(" SA: ");
				serial_print_int_hex_16bit(addr->dest_sa);
				serial_putc(' ');
				break;
		}		
	} else {
		serial_print_str(" (blank) ");
	}
}

void wpan_print_frame_type(uns8 frame_type) {

	switch(frame_type) {
		case FRAME_TYPE_BEACON:
			serial_print_str(" Beacon ");
			break;
		case FRAME_TYPE_DATA:
			serial_print_str(" Data ");
			break;
		case FRAME_TYPE_ACK:
			serial_print_str(" Ack ");
			break;
		case FRAME_TYPE_MAC_COMMAND:
			serial_print_str(" Mac ");
			break;
	}
}

void wpan_print_mac_command(uns8 *data) {

uns8 mac_command = data[0];	// Uses less flash this way

	switch (data[0]) {
		
		case MAC_CMD_ASSOC_REQ:
			serial_print_str("ASSOC REQ (Capability: ");
			serial_print_int_hex(data[1]);
			serial_putc(')');
			break;
		case MAC_CMD_ASSOC_RES:
			serial_print_str("ASSOC RES (Short addr: ");
			serial_print_int_hex(data[2]);	// msb
			serial_print_int_hex(data[1]);	// lsb
			serial_print_str(" Status: ");
			serial_print_int_hex(data[3]);
			serial_putc(')');
			break;
		case MAC_CMD_DISASSOC:
			serial_print_str("DISASSOC (Reason: ");
			serial_print_int_hex(data[1]);
			serial_putc(')');
			break;
		case MAC_CMD_DATA_REQ:
			serial_print_str("DATA REQ");
			break;
		case MAC_CMD_PAN_ID_CONFLICT:
			serial_print_str("PAN ID CONFLICT");
			break;
		case MAC_CMD_ORPHAN:
			serial_print_str("ORPHAN");
			break;
		case MAC_CMD_BEACON_REQ:
			serial_print_str("BEACON REQ");
			break;
		case MAC_CMD_COORD_REALIGN:
			serial_print_str("COORD REALIGN (data TBD)");
			break;
		case MAC_CMD_GTS_REQ:
			serial_print_str("GTS REQ (Char: ");
			serial_print_int_hex(data[1]);
			serial_putc(')');
			break;
	}		
}	


void wpan_process() {
	
	if (pkt_received) {
		wpan_handle_receive();
	}
}

  	



void wpan_handle_receive() {

wpan_address addr;

uns8 temp;
uns8 count;
uns8 frame_type;
uns8 buffer_pos;


	// destination address
	addr.dest_address_type = wpan_rx_buffer[1] >> 6;	// fc lsb dest address type
	
	buffer_pos = 0;
	
	switch (addr.dest_address_type) {
		case WPAN_ADDR_TYPE_EXTENDED: // extended address
	
			// get dest pan
			addr.dest_pan_id = wpan_rx_buffer[4];	// msb
			addr.dest_pan_id <<= 8;
			addr.dest_pan_id += wpan_rx_buffer[3];	// lsb
			// get extended address
			for (count = 0; count < 8; count++) {
				addr.dest_ea[count] = wpan_rx_buffer[5+count];
			}
			// might be faster to unwind this loop...
			buffer_pos = 13;
			break;
		case WPAN_ADDR_TYPE_SHORT: // short address
			// get dest pan
			addr.dest_pan_id = wpan_rx_buffer[4];	// msb
			addr.dest_pan_id <<= 8;
			addr.dest_pan_id += wpan_rx_buffer[3];	// lsb
			// get short address
			addr.dest_sa = wpan_rx_buffer[6];	// msb
			addr.dest_sa <<= 8;
			addr.dest_sa += wpan_rx_buffer[5];	// lsb
			buffer_pos = 7;
			break;
		case WPAN_ADDR_TYPE_NONE: // no address
			buffer_pos = 3;
			break;
	}
	
	// source address
	
	addr.source_address_type = (wpan_rx_buffer[1] >> 2) & 0b00000011;	// fc lsb src address type
	
	switch (addr.source_address_type) {
		case WPAN_ADDR_TYPE_EXTENDED: // extended address
			// get source pan
			if (test_bit(wpan_rx_buffer[0], 6)) {	// pan id compression
				addr.source_pan_id = addr.dest_pan_id;
			} else {	
				addr.source_pan_id = wpan_rx_buffer[buffer_pos+1];	// msb
				addr.source_pan_id <<= 8;
				addr.source_pan_id += wpan_rx_buffer[buffer_pos];	// lsb
				buffer_pos += 2;
			}	
			// get extended address
			for (count = 0; count < 8; count++) {
				addr.source_ea[count] = wpan_rx_buffer[buffer_pos++];
			}

			break;
		case WPAN_ADDR_TYPE_SHORT: // short address
			if (test_bit(wpan_rx_buffer[0], 6)) {	// pan id compression
				addr.source_pan_id = addr.dest_pan_id;
			} else {	
				// get pan
				addr.source_pan_id = wpan_rx_buffer[buffer_pos+1];	// msb
				addr.source_pan_id <<= 8;
				addr.source_pan_id += wpan_rx_buffer[buffer_pos];	// lsb
				buffer_pos += 2;
			}	
			// get short address
			addr.source_sa = wpan_rx_buffer[buffer_pos+1];	// msb
			addr.source_sa <<= 8;
			addr.source_sa += wpan_rx_buffer[buffer_pos];	// lsb
			buffer_pos += 2;
			break;
		case WPAN_ADDR_TYPE_NONE: // no address
			break;
	}
	//debug_var(" buffer pos = ", buffer_pos);
	//debug_var(" wpan_rx_count = ", wpan_rx_count);
	// data should be at buffer_pos
	frame_type = wpan_rx_buffer[0] & 0b00000111;	// frame type
	switch(frame_type) {
		case FRAME_TYPE_MAC_COMMAND:
			//wpan_handle_mac_command(&addr, &wpan_rx_buffer[buffer_pos], wpan_rx_count - buffer_pos);
			break;
		case FRAME_TYPE_BEACON:
			//wpan_handle_beacon(&addr, &wpan_rx_buffer[buffer_pos], wpan_rx_count - buffer_pos);
			break;
		case FRAME_TYPE_DATA:
			wpan_data_received_callback(&addr, &wpan_rx_buffer[buffer_pos], wpan_rx_count - buffer_pos - 4, 
			                            /* lqi */ wpan_rx_buffer[wpan_rx_count-2], /* rssi */ wpan_rx_buffer[wpan_rx_count-1]);
			break;
		// We'll never see an ACK frame, that should be handled by the MRF
	}						   
							   
	pkt_received = 0;
}


void wpan_setup_io() {
	mrf24j40_setup_io();
}

void wpan_init() {
	
	mrf24j40_init();
	
}		
		
