//-----------------------------------------------------
// Pic Pack library
// 
// Embedded Adventures Mega Clock.c
//
// Grab yourself two Sure 6432 boards and create an amazing
// 64x64 clock with temperature and ITS networking capabilities.
//
// Ian Harris 2010
// imharris [at] gmail.com
//
//-----------------------------------------------------
void update_clock();

// Pic Pack includes

// Configuration
#include "config.h"
// Utilities
#include "pic_utils.h"
// Serial port
#include "pic_serial.h"
// Debug routines
#include "debug.h"
// RTC
#include "m41t81s.h"
// MRF24J40 (ieee 802.15.4 support)
#include "mrf24j40.h"
// 802.15.4 (WPAN) PHY/MAC support
#include "wpan.h"
// ITS networking (mode 2)
#include "its_mode2.h"
// ITS networking (general routines)
#include "its_common.h"
// Hardware SPI support
#include "spi_hw.h"
// Tick routines
#include "pic_tick.h"
// Support for Timer 0
#include "pic_timer.h"
// Support for Timer 1
#include "pic_timer1.h"
// Draw library
#include "draw\draw.h"
// Internal screen buffer support
#include "draw\draw_screen_buffer.h" 
// Terminal routines for interaction with the PIC
#include "pic_term.h"
// Temp sensor
#include "tmp75.h"
// Led routines
#include "platform_leds.h"
// System includes
#include "string.h"
#include "stdlib.h"

// Time keeping variables
 
uns8 minutes =59; 
uns8 hours = 11; 
uns8 seconds = 59;
uns8 day = 1;
uns8 date = 1;
uns8 month = 1;
uns8 year = 10;

// Notification that a second has ticked over

bit seconds_trigger = 0;

// temperature

uns16 temperature = 0;
uns16 outside_temp = 0;
bit show_inside = 1;

// Interrupt routine - - - - - - - - - -

void interrupt() {

	// very important for timer 1 handler to be first: It refreshes
	// the display 
	timer_handle_1_isr();
}

void interrupt_low() {	
	
	// handle serial transmit interrupt
	serial_handle_tx_isr();
	// handle serial receive interrupt
	serial_handle_rx_isr();
	// Handle timer 0 interrupt (used for mrf24j40)
	timer_handle_0_isr();

	// Handle mrf24j40 interrupt
	if (test_bit(intcon, INT0IE) && test_bit(intcon, INT0IF)) {
		clear_bit(intcon, INT0IF);
		mrf24j40_handle_isr();
	}
	

	// Handle trigger to update clock every second
	if (test_bit(intcon3, INT1IF)) {
		clear_bit(intcon3, INT1IF);
		seconds_trigger = 1;
	}

}

void timer_1_callback() {
	// refresh display from draw buffer
	drv_refresh();	
}
	


// configure_system
//
// Get the pic ready for everything we want to do

void configure_system() {

	uns8 my_ea[8] = { 0, 0, 0, 0, 0, 0, 0, 5 };	    

	kill_interrupts();	// turn off interrupts just in case

	turn_analog_inputs_off();	// kill those pesky analogue inputs
	
	// serial setup ----------------------------
	serial_setup(SPBRG_115200);	// set serial port up for 115.2k

	term_init();	// initialise terminal 

	// ITS networking setup ----------------------------
	
	// spi interface
	spi_hw_setup_io();
	spi_hw_init();
	
	// interrupt for mrf24j40 INT pin
	clear_bit(intcon, INT0IF);	// in case it was triggered already
	clear_bit(intcon2, INTEDG0);	// need falling edge for mrf24j40 defaults
	set_bit(intcon, INT0IE);	// turn on rb0/int0 interrupts

	// PHY and MAC layer
	mrf24j40_setup_io();
	mrf24j40_init();
	
	mrf24j40_set_channel(ITS_FIXED_CHANNEL);
	mrf24j40_set_extended_address(&my_ea);
	
	// NET layer
	its2_router_init(0x0005, ITS_NETWORK_ID);

	// temp sensor setup ----------------------------

	tmp75_setup_io();
	
	// 12 bit temp, continue temp, shutdown=0
	// OS R1 R0 F1 F0 POL TM SD
	// 0  1  1  0  0  0   0  0   
	tmp75_set_config(TMP75_ADDR, 0b01100000);

	// real time clock setup ----------------------------

	rtc_setup_io();

	rtc_start_clock();	// unpause the m41t81s
	rtc_set_sqw_freq(rtc_sqw_freq_1Hz);	// give us a shove every second
	rtc_start_sqw_output();	// turn on the 1Hz output
	
	make_input(PORTB, 1);	// sqw wired to portb.1
	clear_bit(intcon3, INT1IF);
	set_bit(intcon3, INT1IE);		// enable int1
	
	// draw library setup ----------------------------

	draw_setup_io(); 
	draw_init(); 

	// LED setup ----------------------------
	
	make_output(PORTC, 1);	// portc.1 is our led 
	clear_pin(PORTC, 1);	// make it an output, turn it off

	// general setup ----------------------------

	// Timer 0 is used for ticks - 1ms at 40mhz (10mips)
	timer_setup_0(TIMER_16BIT_MODE, TIMER_PRESCALER_OFF, 0xffff - 10000 -1); 

	// Timer 1 is used for led board refresh

	timer_setup_1(TIMER1_PRESCALER_1_TO_2, 0xffff - 1500 -1);



	turn_peripheral_ints_on();	// turn on peripherals
	turn_global_ints_on();	// turn on interrupts

}


	uns8 EA_1[8] = { 0, 0, 0, 0, 0, 0, 0, 1 };
	uns8 EA_2[8] = { 0, 0, 0, 0, 0, 0, 0, 2 };
	uns8 EA_3[8] = { 0, 0, 0, 0, 0, 0, 0, 3 };

void term_entry_callback(uns8 *term_buffer) {

uns8 data[5];
uns8 first_char;
	
	first_char = term_buffer[0];
	switch (first_char) {
	case '1':
		serial_print_str("TX1... ");
		data[0] = 'T';
		data[1] = 'O';
		// we just happen to know the device is 0x2020
		its2_router_queue_packet(0x0001, ITS_GENERIC_DATA, &data, 2, 1);

		break;
	case '2':
		serial_print_str("TX2.... ");
		data[0] = 'o';
		data[1] = 't';
		its2_router_queue_packet(0x0002, ITS_GENERIC_DATA, &data, 2, 1);
		
		break;
	case '3':
		serial_print_str("TX3.... ");
		data[0] = 'T';
		data[1] = 't';
		its2_router_queue_packet(0x0003, ITS_GENERIC_DATA, &data, 2, 1);
		
		break;
	case 'a':
		serial_print_str("Setting address to 1\n");
		mrf24j40_set_extended_address(&EA_1);
		its2_router_init(0x0001, ITS_NETWORK_ID);
		
		break;
	case 'b':
		serial_print_str("Setting address to 2\n");
		mrf24j40_set_extended_address(&EA_2);
		its2_router_init(0x0002, ITS_NETWORK_ID);
		break;
		
	case 'c':
		serial_print_str("Setting address to 3\n");
		mrf24j40_set_extended_address(&EA_3);
		its2_router_init(0x0003, ITS_NETWORK_ID);
		break;
	case 'd':
		serial_print_str("Setting pan id = 5\n");
		mrf24j40_set_pan_id(0x05);
		break;
	case 'u':
		update_clock();
		break;
	case 'p':
			its2_print_queue();
			break;
		
	} // switch		

}

void its2_router_receive_callback(uns16 device_id, uns8 *data, uns8 data_length) {

//		flash_led(1);

		serial_print_str("Controller receives from ");
		serial_print_int(device_id);
		serial_print_str(" bytes=");
		serial_print_int(data_length);
		serial_print_str(" data = ");
		outside_temp = data[1];
		outside_temp <<= 8;
		outside_temp = outside_temp + data[0];
		serial_print_str("OUTSIDETEMP=");
		serial_print_int(outside_temp);
}


void its2_transmit_status_callback(its2_packet *pkt, uns8 status) {
	serial_print_str("{TX src:");
	serial_print_int(pkt->its_source_id);
	serial_print_str(" dst: ");
	serial_print_int(pkt->its_dest_id);
	serial_print_str(" stat=");
	switch (status) {
		case ITS2_TX_STATUS_SUCCESS:
			serial_print_str("Success");
			break;
		case ITS2_TX_STATUS_NO_ROUTE:
			serial_print_str("No route");
			break;
		case ITS2_TX_STATUS_NO_ACK:
			serial_print_str("No ack");
			break;
		case ITS2_TX_STATUS_NEXT_HOP_UNKNOWN:
			serial_print_str("Next hop unknown");
			break;
	}
	serial_print_str("} ");
	
	if (status == ITS2_TX_STATUS_SUCCESS) {
		platform_leds_flash(2);
	} else {
		platform_leds_flash(3);
	}	
}	
 
void clock_tick(uns8 x, uns8 y, uns8 col) {
	draw_set_pixel(x, y, col);
	draw_set_pixel(x-1, y, 3);
	draw_set_pixel(x+1, y, 3);
	draw_set_pixel(x, y-1, 3);
	draw_set_pixel(x, y+1, 3);
}	 
 
 
void clock_number_ticks() { 
		clock_tick(61, 32,2); 
		clock_tick(57,	46,2); 
		clock_tick(46,	57,2); 
		clock_tick(31,	61,2); 
		clock_tick(17,	57,2); 
		clock_tick(6,	46,2); 
		clock_tick(2,	31,2); 
		clock_tick(6,	17,2); 
		clock_tick(17,	6,2); 
		clock_tick(32,	2,2); 
		clock_tick(46,	6,2); 
		clock_tick(57,	17,2); 
 
		//draw_set_pixel(31,31, 1); 
		//draw_set_pixel(32,31, 1); 
		//draw_set_pixel(31,32, 1); 
		//draw_set_pixel(32,32, 1); 
} 
 
 
rom char *minute_hand_y = { 
 4, 4, 4, 5, 6, 7, 9,11,13,16,18,20,23,26,29,
32,34,37,40,43,45,47,50,52,54,56,57,58,59,59,
59,59,59,58,57,56,54,52,50,47,45,42,40,37,34,
31,29,26,23,21,18,16,13,11, 9, 7, 6, 5, 4, 4}; 
 
rom char *minute_hand_x = { 
32,34,37,40,43,45,47,50,52,54,56,57,58,59,59,
59,59,59,58,57,56,54,52,50,47,45,43,40,37,34,
31,29,26,23,20,18,16,13,11, 9, 7, 6, 5, 4, 4,
 4, 4, 4, 5, 6, 7,9,11,13,16,18,21,23,26,29,}; 
 
 
 
 
 
rom char *hour_hand_y = { 
12	,12	,12	,13	,14	,14	,16	,17	,19	,21	,23	,24	,26	,28	,30	,
32	,34	,36	,38	,40	,42	,44	,46	,47	,49	,49	,50	,50	,51	,51	,
51	,51	,51	,50	,50	,49	,47	,46	,44	,42	,40	,39	,37	,35	,33	,
31	,29	,27	,25	,22	,21	,19	,17	,16	,14	,13	,13	,13	,12	,12	,}; 
 
rom char *hour_hand_x = { 
32	,34	,36	,38	,40	,41	,43	,45	,47	,48	,49	,50	,50	,51	,51	,
51	,51	,51	,50	,49	,48	,47	,45	,43	,41	,41	,39	,37	,35	,33	,
31	,29	,27	,26	,24	,22	,20	,18	,16	,15	,14	,13	,13	,12	,12	,
12	,12	,12	,13	,14	,15	,16	,18	,20	,22	,24	,25	,26	,28	,30	,}; 
 
 
 
 
 
rom char *second_hand_y = { 
2	,2	,3	,3	,5	,6	,8	,9	,11	,14	,17	,19	,22	,25	,29	,
32	,35	,38	,41	,44	,46	,49	,52	,54	,55	,57	,58	,60	,60	,61	,
61	,61	,60	,60	,58	,57	,55	,53	,51	,49	,46	,44	,41	,38	,34	,
31	,29	,25	,22	,19	,17	,14	,11	,9	,8	,6	,5	,3	,3	,2	,
};

rom char *second_hand_x = {
32	,35	,38	,41	,44	,46	,49	,51	,53	,55	,57	,58	,60	,60	,61	,
61	,61	,60	,60	,58	,57	,55	,53	,51	,49	,46	,44	,41	,38	,35	,
31	,28	,25	,22	,19	,17	,14	,11	,9	,8	,6	,5	,3	,3	,2	,
2	,2	,3	,3	,5	,6	,8	,10	,12	,14	,17	,19	,22	,25	,28	,
};
	
draw_test_hands() {
	uns8 count;
	
	for (count=0;count<60;count++) {
		draw_set_pixel(second_hand_x[count], second_hand_y[count],3);
		//draw_set_pixel(minute_hand_x[count], minute_hand_y[count],3);
		//draw_set_pixel(hour_hand_x[count],   hour_hand_y[count], 3);
	}
		
}	

void clock_second_hand(uns8 seconds) {
	uns8 startx, starty;
	
	if (seconds >= 45) { 	
		startx = 31; 
		starty = 31;
	} else if (seconds >= 30) { 
		startx = 31; 
		starty = 32;
	} else if (seconds >= 15) { 
		startx = 32; 
		starty = 32;
	} else { 
		startx = 32; 
		starty = 31; 
	} 

	draw_line(startx, starty, second_hand_x[seconds], second_hand_y[seconds], 1); 
}	
		 
void clock_minute_hand(uns8 minutes) { 
	uns8 startx, starty;
	
	if (minutes < 15) { 
		startx = 32; 
		starty = 31;
	} else if (minutes < 30) { 
		startx = 32; 
		starty = 32;
	} else if (minutes < 45) { 
		startx = 31; 
		starty = 32;
	} else { 
		startx = 31; 
		starty = 31; 
	} 
	 
	draw_line(startx, starty, minute_hand_x[minutes], minute_hand_y[minutes], 3); 
		 
} 
	 
void clock_hour_hand(uns8 hours, uns8 minutes) { 
	uns8 startx, starty, endx, endy; 
	uns8 index, lookup; 
	if (hours > 12) {
		hours = hours - 12;
	}	 
	index = hours * 5 + minutes / 12;
	 
	if (index < 15) { 
		startx = 32; 
		starty = 31;
	} else if (index < 30) { 
		startx = 32; 
		starty = 32;
	} else if (index < 45) { 
		startx = 31; 
		starty = 32;
	} else { 
		startx = 31; 
		starty = 31;
	} 
	draw_line(startx, starty, hour_hand_x[index], hour_hand_y[index], 2); 
	 
} 
 
void clock_date(uns8 date) {

char buffer[3];
	
	if (date > 9) {
		buffer[0] = '0' + date / 10;
		buffer[1] = '0' + date % 10;
	} else {
		buffer[0] = '0';
		buffer[1] = '0' + date;
	}
	buffer[2] = '\0';
	draw_print_str(0,  63, 11, 0, 3, buffer);
}

char *day_string = "Mo\0Tu\0We\0Th\0Fr\0Sa\0Su";
	 
void clock_day(uns8 day) {
	
	uns16 length;
	day = day - 1;
	day = day * 3;
	length = draw_length_str(&day_string[day]);
	draw_print_str(64-length, 6, 11, 0, 3, &day_string[day]);
}

char *month_string = "Ja\0Fb\0Mr\0Ap\0Ma\0Ju\0Jl\0Au\0Se\0Oc\0No\0De\0";

void clock_month(uns8 month) {
	uns16 length;

	if (month > 12) {
		return;
	}	
	month = month - 1;
	month = month * 3;
	
	length = draw_length_str(&month_string[month]);


    draw_print_str(64-length, 63, 16, 0, 3, &month_string[month]);	
}
 
void clock_temperature(uns16 temp, bit inside) {

char buffer[5];
uns8 count;
uns8 temp_int;
bit negative = 0;
	temp_int = temp >> 8; 

	if (temp.15 == 1) {	// negative
		negative = 1;
		temp = temp | 0x0f;	// Set lower 4 bits
		temp = ~temp;	// Take complement
		temp = temp + 1;		// Add one
	}

	temp_int = temp >> 8;
	count = 0;

	if (negative) {
		buffer[0] = '-';
		count = 1;
	}
	
	if (temp_int > 9) {
		buffer[count++] = '0' + temp_int / 10;
		buffer[count++] = '0' + temp_int % 10;
	} else {
		buffer[count++] = '0' + temp_int;
	}
	buffer[count++] = 127; // degree sign
	buffer[count++] = '\0';
	if (inside) {
		draw_print_str(0, 45, 16, 0, 3, buffer);
	} else {
		draw_print_str(0, 45, 16, 0, 2, buffer);
	}	
}

void update_clock() {

	
	// grab the latest time from the RTC chip

	minutes = rtc_get_minutes();
	hours   = rtc_get_hours();
	seconds = rtc_get_seconds();
	day 	= rtc_get_dow();
	date 	= rtc_get_date();
	month 	= rtc_get_month();
	year 	= rtc_get_year();

	// start with a blank slate

	draw_clear_screen();

	// draw what we want on the display

	draw_circle2(31, 31, 31, 1); // outside
	
	clock_date(date);
	clock_day(day);
	clock_month(month);
	clock_minute_hand(minutes); 
	clock_hour_hand(hours, minutes); 

	// update the temperature every now and then
	if (seconds % 5 == 0) {	// every 5 seconds
		temperature = tmp75_get_temp(TMP75_ADDR);	// update temperature
			// request outside temp
		if (show_inside) {
			show_inside = 0;
		} else {
			show_inside = 1;
		}
	}
		its2_router_queue_packet(0x0002, ITS_GENERIC_DATA, NULL, 0, 1);
	
	if (show_inside) {
		clock_temperature(temperature, 1);
	} else {
		clock_temperature(outside_temp, 0);
	}	
	clock_number_ticks();
	clock_second_hand(seconds);
	
	// Paint buffer to display
	draw_paint(); 
}

void main() {

	uns8 my_ea[8] = { 0, 0, 0, 0, 0, 0, 0, 4};
	
	configure_system();

	delay_ms(100);
	
	serial_print_str("\n\nEA Mega Clock\n"); 
	serial_print_str( "\n<"); 
	serial_print_str(__TIME__); 
	serial_putc(' '); 
	serial_print_str(__DATE__); 
	serial_print_str(">\n"); 

	draw_clear_screen(); 
	draw_paint();
	
	timer_start_0();
	timer_start_1();
	rtc_set_month(12);
	rtc_set_date(31);

	for(;;) {	// for ever
		
		term_process();	// process any typing on the terminal
		
		if (seconds_trigger) {
			update_clock();
			seconds_trigger = 0;
		}


		wpan_process();	

		its2_router_process();

	}	
	
}	// main

